/*
 * "" ̕]
 *
 * Copyright (C) 1990 by cron, hirofumi
 * Copyright (C) 1996 by Hirofumi YAMAMOTO, Hiro Software Factory
 */
%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <math.h>

#include "expr.h"

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

double eval_val;
BOOL no_err;
BOOL var_assigned = FALSE;
double acc;

static char rcs_id[] =
	"$Id: expr.y 1.6 1996/05/26 17:55:28 hirofumi Rel hirofumi $";

static int eval_end;
/*static*/ char *eval_s = 0;

int yyerror(char *s);
int	yylex(void);

#if 1

#define mod(a,b)	(fmod((a),(b)))
#define and(a,b)	(double)((long)(a) & (long)(b))
#define or(a,b)		(double)((long)(a) | (long)(b))
#define xor(a,b)	(double)((long)(a) ^ (long)(b))
#define power(a,b)	(pow((a),(b)))
#define shl(a,b)	(double)((long)(a) << (long)(b))
#define shr(a,b)	(double)((long)(a) >> (long)(b))
#define not(a)		(double)(~(long)(a))

#else	/* RTYPE != double */

#define mod(a,b)	((a) % (b))
#define and(a,b)	((a) & (b))
#define or(a,b)		((a) | (b))
#define xor(a,b)	((a) ^ (b))
#define shl(a,b)	((a) << (b))
#define shr(a,b)	((a) >> (b))
#define not(a)		(~(a))

double power(double a, double b)
{
	/* skimp.... */
	double l = a, r;
	
	if (b < 0) {
		return 0;
	}else if (b == 0) {
		return 1;
	}else if (a == 0) {
		if (b >= 0) {
			return 0;
		}
		yyerror("pow ?");
	}else{
		for (r = 1; r < b; ++r) {
			l *= a;
		}
	}
	return l;
}

#endif	/* end of double != double */

double preinc(char* name)
{
	return assign_variable(name, eval_variable(name) + 1);
}

double predec(char* name)
{
	return assign_variable(name, eval_variable(name) - 1);
}

double postinc(char* name)
{
	return acc = assign_variable(name, eval_variable(name) + 1) - 1;
}

double postdec(char* name)
{
	return acc = assign_variable(name, eval_variable(name) - 1) + 1;
}

int yyerror(char *s)
{
	no_err = FALSE;
	//fprintf(stderr, "%s\n", s);
	HandleError(s);
	return -1;
}

%}

%union {
	char* name;
	double value;
}

%token	<name>	VARIABLE

%type <value>	expr
%type <name>	VARIABLE

%token	Y_NUMBER EOL
%token	'+' '-' '*' '/' '%' '(' ')' '=' '&' '|' '<' '>'
%token	PLEQ MIEQ MULEQ DIVEQ MODEQ
%token	INC DEC POWER
%token	SHL SHR

%left	','
%right	'=' PLEQ MIEQ MULEQ DIVEQ MODEQ ANDEQ XOREQ OREQ
%right	'?' ':'
%left	OR
%left	AND
%left	IDN DIF '<' LEQ '>' GEQ
%left	'|'
%left	'^'
%left	'&'
%left	'+' '-'
%left	'*' '/' '%'
%left	SHL SHR
%left	POWER
%nonassoc	UNARY INC DEC '!' '~'

%%
line	: /* empty */
		| line EOL
		| line expr EOL			{ 
				eval_val = $2; 
			}
		| line error EOL		{ yyerrok; }
		;


expr	:expr ',' expr			{ var_assigned = FALSE; $$ = $3; }
		| '-' expr	%prec UNARY	{ $$ = -$2; }
		| '+' expr	%prec UNARY	{ $$ = $2; }

		| INC VARIABLE			{ $$ = preinc($2); }
		| DEC VARIABLE			{ $$ = predec($2); }
		| VARIABLE INC			{ $$ = postinc($1); }
		| VARIABLE DEC			{ $$ = postdec($1); }

		| '!' expr				{ $$ = !($2); }
		| '~' expr				{ $$ = not($2); }

		| expr '+' expr			{ $$ = $1 + $3; }
		| expr '-' expr			{ $$ = $1 - $3; }
		| expr '*' expr			{ $$ = $1 * $3; }
		| expr '/' expr			{ $$ = $1 / $3; }
		| expr '%' expr			{ $$ = mod($1, $3); }
		| expr '&' expr			{ $$ = and($1, $3); }
		| expr '|' expr			{ $$ = or($1, $3); }
		| expr '^' expr			{ $$ = xor($1, $3); }
		| expr SHL expr			{ $$ = shl($1, $3); }
		| expr SHR expr			{ $$ = shr($1, $3); }

		| expr POWER expr		{ $$ = power($1, $3); }

		| expr AND expr			{ $$ = $1 && $3; }
		| expr OR expr			{ $$ = $1 || $3; }
		| expr IDN expr			{ $$ = $1 == $3; }
		| expr DIF expr			{ $$ = $1 != $3; }
		| expr '<' expr			{ $$ = $1 < $3; }
		| expr LEQ expr			{ $$ = $1 <= $3; }
		| expr '>' expr			{ $$ = $1 > $3; }
		| expr GEQ expr			{ $$ = $1 >= $3; }
		| expr '?' expr ':' expr	{ 
				if ($1)
					$$ = $3;
				else
					$$ = $5;
			}
		| VARIABLE '=' expr {
				$$ = assign_variable($1, $3);
			}
		| VARIABLE PLEQ expr {
				$$ = assign_variable($1, eval_variable($1) + $3);
			}
		| VARIABLE MIEQ expr {
				$$ = assign_variable($1, eval_variable($1) - $3);
			}
		| VARIABLE MULEQ expr {
				$$ = assign_variable($1, eval_variable($1) * $3);
			}
		| VARIABLE DIVEQ expr {
				$$ = assign_variable($1, eval_variable($1) / $3);
			}
		| VARIABLE MODEQ expr {
				$$ = assign_variable($1, mod(eval_variable($1), $3));
			}
		| VARIABLE ANDEQ expr {
				$$ = assign_variable($1, and(eval_variable($1), $3));
			}
		| VARIABLE XOREQ expr {
				$$ = assign_variable($1, xor(eval_variable($1), $3));
			}
		| VARIABLE OREQ expr {
				$$ = assign_variable($1, or(eval_variable($1), $3));
			}
		| VARIABLE {
				$$ = eval_variable($1);
			}
		| Y_NUMBER { }
		| '(' expr ')' {
				 $$ = $2;
		}
		;

%%

char* atoh(char* s, double* r)
{
	register short c;

	*r = (double)0;
	while (isxdigit(c = *s)) {
		*r *= 0x10;
		*r += isdigit(c) ? c - '0' : tolower(c) - 'a' + 0xa;
		++s;
	}
	return s;
}

#define plural(c,fc,sc,id)		\
	if ((c) == (fc)) {	\
		if (((c) = *eval_s++) == (sc))	{\
			printf("detected %c%c\n", (fc), (sc)); \
			return id;	\
		} \
		else {	\
			--eval_s;	\
			printf("detected %c\n", (fc)); \
			return fc;	\
		}	\
	}
#define plural3(c,fc,sc,id1,tc,id2) \
	if ((c) == (fc)) { \
		if (((c) = *eval_s++) == (sc)) \
			return id1; \
		else if ((c) == (tc)) \
			return id2; \
		else { \
			--eval_s; \
			return fc; \
		} \
	}

int	yylex(void)
{
	if (eval_end)
		//return YYEOF;
		return EOF;
	else {
		int	c;
		do {
			c = *eval_s++;
			if (isalpha(c) || c == '_') {
				static char buff[1024];		// ꂾƂƂΑv낤
				char *p = buff;
				yylval.name = buff;
				eval_s--;
				do {
					*p++ = *eval_s++;
				} while (isalnum(*eval_s) || *eval_s == '_');
				*p = '\0';
				yylval.name = xstrdup(buff);
				return VARIABLE;
			}else if (isdigit(c)) {
				if (c == '0' && tolower(*eval_s) == 'x') {
					++eval_s;
					/* skimp.... */
					eval_s = atoh(eval_s, &yylval.value);
				}else{
					yylval.value = atof(--eval_s);
				}

				{
					BOOL point = FALSE;
					BOOL exp = FALSE;
					#define FORM_ERR	"_̃tH[}bgႢ܂"
					for (; isdigit(*eval_s); ++eval_s) {
						if (eval_s[1] == '.') {
							++eval_s;
							if (point || exp) {
								yyerror(FORM_ERR);
							}else{
								point = TRUE;
							}
						}else if (tolower(eval_s[1]) == 'e') {
							eval_s += 2;
							if (exp) {
								yyerror(FORM_ERR);
							}else{
								exp = TRUE;
								if (*eval_s != '+' && *eval_s != '-') {
									yyerror(FORM_ERR);
								}
							}
						}
					}
				}
				return Y_NUMBER;
			}
			else plural3(c, '*', '*', POWER, '=', MULEQ)
			else plural3(c, '<', '<', SHL, '=', LEQ)
			else plural3(c, '>', '>', SHR, '=', GEQ)
			else plural3(c, '+', '+', INC, '=', PLEQ)
			else plural3(c, '-', '-', DEC, '=', MIEQ)
			else plural3(c, '&', '&', AND, '=', ANDEQ)
			else plural3(c, '|', '|', OR, '=', OREQ)
			else plural(c, '=', '=', IDN)
			else plural(c, '!', '=', DIF)
			else plural(c, '/', '=', DIVEQ)
			else plural(c, '%', '=', MODEQ)
			else plural(c, '^', '=', XOREQ)
			else if (c == '\0') {
				eval_end = 1;
				return EOL;
			}
		} while (isspace(c)) ;
		return c;
	}
}

double* evaluate(char *s)
{
	extern int yyparse();

	acc = (double)0;
	var_assigned = FALSE;
	no_err = TRUE;
	if (s == NULL)
		return NULL;
	eval_end = 0;
	eval_s = s;
	yyparse();
	xstrfree();
	acc = eval_val;
	if (!no_err)
		return NULL;
	return &acc;
}
