/////////////////////////////////////////////////////////////////////////////
// ϐ
//
// Copyright (C) 1996 by Hirofumi YAMAMOTO, Hiro Software Factory. All rights reserved.
//

#include "stdafx.h"
#include "ecalc.h"
#include "var.h"
#include "expr.h"

static char rcs_id[] =
	"$Id: var.cpp 1.4 1996/05/26 17:06:04 hirofumi Rel $";

#define new	DEBUG_NEW

Variables g_variables;

/////////////////////////////////////////////////////////////////////////////
// CStringMapNX
BOOL CStringMap::operator[](LPCTSTR s)
{
	void* v;
	if (Lookup(s, v))
		return TRUE;
	return FALSE;
}

void CStringMap::Add(LPCTSTR s)
{
	SetAt(s, NULL);
}

/////////////////////////////////////////////////////////////////////////////
// Zwp

void HandleError(const char* msg)
{
	if (!strcmp(msg, "syntax error"))
		msg = "@Ⴂ܂";
	CString tmp(msg);
	MessageBeep(MB_ICONINFORMATION);
	AfxMessageBox(msg, MB_ICONINFORMATION | MB_OK);
}

double eval_variable(const char* name)
{
	CString nn(name);	// MBCS,Wide Char̖ʓ|Ă炤
	return g_variables.GetVariable(nn);
}

double assign_variable(const char* name, double value)
{
	CString nn(name);
	g_variables.SetVariable(nn, value);
	return value;
}


char* xstrdup(const char* s)
{
	return g_variables.AllocWorkStr(s);
}

void xstrfree()
{
	g_variables.DeleteAllWorkStr();
}

/////////////////////////////////////////////////////////////////////////////
// Variable NX

Variable::Variable(BOOL isReserved)
	: m_isReserved(isReserved), m_type(T_VALUE)
{
}

Variable::~Variable()
{
	m_lookup.RemoveAll();
}

void Variable::DumpBin(CString& r, __int64 value, int col/* = 0 */)
{
	if (value > 1) {
		DumpBin(r, value >> 1, col + 1);
		if (col && col % 4 == 0)
			r += " ";
	}
	if (value & 1)
		r += "1";
	else
		r += "0";
}

void Variable::SetValue(double value)
{
	m_value = value;
	m_type = T_VALUE;
	m_expr = _T("");
	m_lookup.RemoveAll();
}

Variable::operator double()
{
	return m_value;
}

CString Variable::GetValue(int format)
{
	CString r;
	if (format == DF_DEC) {
		r.Format(_T("%.15g"), m_value);
	}
	else if (format == DF_BIN) {
		// NOTICE: __int64́AMicrosoftŗL̊g
		__int64 i64 = (__int64)m_value;
		DumpBin(r, i64);
	}
	else if (format == DF_HEX) {
		__int64 i64 = (__int64)m_value;
		LPTSTR p = r.GetBuffer(64);
		sprintf(p, _T("0x%I64X"), i64);
		r.ReleaseBuffer();
	}
	else if (format == DF_EXPR)
		return GetExpression();
	else {
		ASSERT(0);
	}
	return r;
}

BOOL Variable::IsReserved()
{
	return m_isReserved;
}

int Variable::GetListIndex()
{
	if (m_type == T_VALUE)
		return !IsReserved();
	return 3;	// 
}

void Variable::SetExpression(LPCTSTR expr)
{
	m_type = T_EXPR;
	m_expr = expr;
}

const CString& Variable::GetExpression()
{
	return m_expr;
}

void Variable::InitCyclicCheck()
{
	m_referred = m_calculated = FALSE;
	m_lookup.RemoveAll();
}


/////////////////////////////////////////////////////////////////////////////
// ConstantNX

Constant::Constant(double val)
	: Variable(TRUE)
{
	m_value = val;
}

void Constant::SetValue(double)
{
	AfxMessageBox(IDP_CONSTANT_CANT_BE_ASSIGNED);
}

int Constant::GetListIndex()
{
	return 2;
}

/////////////////////////////////////////////////////////////////////////////
// Variables RNVNX

Variables::Variables()
{
	// \ϐĂ
	Variable* v = new Variable(TRUE);	// \
	v->SetValue(0.0);
	SetAt(resultVarName, v);
	v = new Constant(3.141592653589793);
	SetAt(_T("pi"), v);
}

Variables::~Variables()
{
	// ϐ}bv̍폜
	POSITION pos = GetStartPosition();
	while (pos) {
		CString name;
		Variable* v = NULL;
		GetNextAssoc(pos, name, v);
		delete v;
	}
	RemoveAll();
}

#define REG_VAR	_T("variables-deafult")
#define VAR_SUMMARY	_T("00-summary")
#define TYPE_NAME	_T("type-var-%u")
#define VAR_NAME	_T("name-var-%u")
#define VAR_VALUE	_T("var-%u")

void Variables::ReadReg()
{
	CWinApp* app = AfxGetApp();
	ASSERT(app);
	UINT upto = app->GetProfileInt(REG_VAR, VAR_SUMMARY, 0);
	for (UINT i = 0; i < upto; ++i) {
		CString name;
		name.Format(VAR_NAME, i);
		CString nn = app->GetProfileString(REG_VAR, name);
		if (nn.IsEmpty())
			continue;
		Variable* v = new Variable();
		name.Format(TYPE_NAME, i);
		UINT type = app->GetProfileInt(REG_VAR, name, Variable::T_VALUE);
		name.Format(VAR_VALUE, i);
		if (type == Variable::T_VALUE) {
			LPBYTE lpb = NULL;
			double d = 0.0;
			UINT bytes = sizeof d;
			if (!app->GetProfileBinary(REG_VAR, name, &lpb, &bytes)) {
				delete v;
				continue;
			}
			//ASSERT((void*)lpb == (void*)&d);	// lpbɓnlɂ͈ӖȂ
			ASSERT(lpb);
			ASSERT(bytes == sizeof d);
			memcpy(&d, lpb, bytes);
			delete lpb;
			v->SetValue(d);
		}
		else if (type == Variable::T_EXPR) {
			CString expr = app->GetProfileString(REG_VAR, name);
			v->SetExpression(expr);
		}
		else {
			ASSERT(0);
		}
		ASSERT((*this)[nn] == NULL);
		SetAt(nn, v);
	}
	CreateTree();
	Calculate();
}

void Variables::WriteReg()
{
	int upto = GetCount();
	CWinApp* app = AfxGetApp();
	POSITION pos = GetStartPosition();
	int i = 0;
	while (pos) {
		CString nn;
		Variable* v = NULL;
		GetNextAssoc(pos, nn, v);
		ASSERT(v);
		if (!v->IsReserved()) {
			CString name;
			name.Format(TYPE_NAME,  i);
			app->WriteProfileInt(REG_VAR, name, v->m_type);
			name.Format(VAR_NAME, i);
			app->WriteProfileString(REG_VAR, name, nn);
			name.Format(VAR_VALUE, i);
			if (v->m_type == Variable::T_VALUE) {
				app->WriteProfileBinary(REG_VAR, name, (LPBYTE)&v->m_value, sizeof v->m_value);
			}
			else 
				app->WriteProfileString(REG_VAR, name, (LPCTSTR)v->GetExpression());
			++i;
		}
	}
	app->WriteProfileInt(REG_VAR, _T(VAR_SUMMARY), i);
}


Variable* Variables::operator[](LPCTSTR name)
{
	Variable* v = NULL;
	Lookup(name, v);
	return v;
}

double Variables::GetVariable(LPCTSTR name)
{
	if (m_checkingCyclic) {
		AddLookup(name);
	}
	Variable* v = (*this)[name];
	if (v)
		return *v;
	return 0.0;
}

void Variables::SetVariable(LPCTSTR name, double value)
{
	if (m_checkingCyclic) {
		AfxMessageBox(IDP_CANT_ASSIGN_VAR_IN_EXPR);
		throw -1;
	}
	Variable* v = g_variables[name];
	if (v)
		v->SetValue(value);
	else {
		v = new Variable();
		v->SetValue(value);
		g_variables.SetAt(name, v);
	}
	//RefreshVariableDisplay();
}

char* Variables::AllocWorkStr(LPCTSTR str)
{
	int n = m_nameWork.Add(str);
	return m_nameWork[n].GetBuffer(0);
}

void Variables::DeleteAllWorkStr()
{
	m_nameWork.RemoveAll();
}

void Variables::CalcEachVar(Variable* v)
{
	ASSERT(v);
	if (v->IsCalculated())
		return;
	v->SetReferred();
	v->SetCalculated();
	POSITION pos = v->m_lookup.GetStartPosition();
	while (pos) {
		CString nn;
		void* p = NULL;
		v->m_lookup.GetNextAssoc(pos, nn, p);
		ASSERT(!nn.IsEmpty());
		Variable* vv = (*this)[nn];
		if (vv)
			CalcEachVar(vv);
	}
	CString tmp = v->GetExpression();
	double* dp = evaluate(tmp.GetBuffer(0));
	ASSERT(dp);
	tmp.ReleaseBuffer();
	v->m_value = *dp;
}

void Variables::Calculate()
{
	// 
	POSITION pos = GetStartPosition();
	while (pos) {
		CString nn;
		Variable* vv = NULL;
		GetNextAssoc(pos, nn, vv);
		ASSERT(vv);
		vv->ClearCalculatedFlag();
	}
	pos = GetStartPosition();
	while (pos) {
		CString nn;
		Variable* v= NULL;
		GetNextAssoc(pos, nn, v);
		ASSERT(v != NULL);
		if (v->IsCalculated())
			continue;
		if (v)
			CalcEachVar(v);
	}
}

void Variables::AddLookup(LPCTSTR name)
{
	ASSERT(name);
	ASSERT(m_currentChecking);
	m_currentChecking->m_lookup.Add(name);
}

void Variables::ClearAllReferenceFlags()
{
	POSITION pos = GetStartPosition();
	while (pos) {
		CString name;
		Variable* v = NULL;
		GetNextAssoc(pos, name, v);
		ASSERT(v);
		v->ClearReferenceFlag();
	}
}

void Variables::CheckEachVar(Variable* v)
{
	ASSERT(v);
	if (v->IsReferred()) {
		AfxMessageBox(IDP_CYCLIC);
		throw -1;
	}
	if (v->IsCalculated())
		return;
	// gĂϐo^
	m_currentChecking = v;
	CString tmp = v->GetExpression();
	if (evaluate(tmp.GetBuffer(0)) == NULL)
		throw -1;
	tmp.ReleaseBuffer();
	v->SetReferred();
	v->SetCalculated();
	POSITION pos = v->m_lookup.GetStartPosition();
	while (pos) {
		CString nn;
		void* p = NULL;
		v->m_lookup.GetNextAssoc(pos, nn, p);
		ASSERT(!nn.IsEmpty());
		Variable* vv = (*this)[nn];
		if (vv)
			CheckEachVar(vv);
	}
}

void Variables::CreateTree()
{
	try {
		m_checkingCyclic = TRUE;
		// 
		POSITION pos = GetStartPosition();
		while (pos) {
			CString nn;
			Variable* vv = NULL;
			GetNextAssoc(pos, nn, vv);
			ASSERT(vv);
			vv->InitCyclicCheck();
		}
		pos = GetStartPosition();
		while (pos) {
			CString nn;
			Variable* vv = NULL;
			GetNextAssoc(pos, nn, vv);
			ASSERT(vv);
			if (vv->IsCalculated())
				continue;
			ClearAllReferenceFlags();
			CheckEachVar(vv);
		}
	}
	catch (...) {
		m_checkingCyclic = FALSE;
		m_currentChecking = NULL;
		throw;
	}
	m_checkingCyclic = FALSE;
	m_currentChecking = NULL;
}

BOOL Variables::CheckCyclic(LPCTSTR name, LPCTSTR expr)
{
	Variable* old = (*this)[name];
	Variable* v = new Variable();
	v->SetExpression(expr);
	CString sname(name);
	SetAt(name, v);
	BOOL r = TRUE;
	try {
		CreateTree();
	}
	catch (...) {
		r = FALSE;
	}
	if (old) {
		SetAt(sname, old);
	}
	else {
		RemoveKey(sname);
	}
	delete v;
	m_currentChecking = NULL;
	return r;
}
